@@ -2,6 +2,8 @@
  * This file is part of jdupes; see jdupes.c for license information */
 
 #include <stdio.h>
+#include <stdint.h>
+#include <inttypes.h>
 #include "jdupes.h"
 #include "act_summarize.h"
 
@@ -31,9 +33,9 @@ extern void summarizematches(const file_t * restrict files)
   else
   {
     printf("%d duplicate files (in %d sets), occupying ", numfiles, numsets);
-    if (numbytes < 1000) printf("%jd byte%c\n", (intmax_t)numbytes, (numbytes != 1) ? 's' : ' ');
-    else if (numbytes <= 1000000) printf("%jd KB\n", (intmax_t)(numbytes / 1000));
-    else printf("%jd MB\n", (intmax_t)(numbytes / 1000000));
+    if (numbytes < 1000) printf("%" PRIdMAX " byte%c\n", (intmax_t)numbytes, (numbytes != 1) ? 's' : ' ');
+    else if (numbytes <= 1000000) printf("%" PRIdMAX " KB\n", (intmax_t)(numbytes / 1000));
+    else printf("%" PRIdMAX " MB\n", (intmax_t)(numbytes / 1000000));
   }
   return;
 }
