@@ -1,5 +1,5 @@
-/* jdupes (C) 2015-2019 Jody Bruchon <jody@jodybruchon.com>
-   Derived from fdupes (C) 1999-2019 Adrian Lopez
+/* jdupes (C) 2015-2020 Jody Bruchon <jody@jodybruchon.com>
+   Forked from fdupes 1.51 (C) 1999-2014 Adrian Lopez
 
    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation files
@@ -21,7 +21,7 @@
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. */
 
 #define VER "1.13.1"
-#define VERDATE "2019-06-10"
+#define VERDATE "2020-06-10"
 
 #include <dirent.h>
 #include <errno.h>
@@ -40,12 +40,32 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <unistd.h>
-/* Optional btrfs support */
-#ifdef ENABLE_BTRFS
- #include <linux/btrfs.h>
+/* Optional FIDEDUPERANGE support */
+#ifdef ENABLE_DEDUPE
+ #ifndef __linux__
+  #error "Filesystem-managed deduplication only available for Linux."
+ #endif /* __linux__ */
  #include <sys/ioctl.h>
  #include <sys/utsname.h>
-#endif
+ #ifdef STATIC_BTRFS_H
+  /* Static BTRFS header */
+  #include <linux/types.h>
+  #define BTRFS_IOCTL_MAGIC 0x94
+  #define BTRFS_DEVICE_PATH_NAME_MAX 1024
+  #define FILE_DEDUPE_RANGE_DIFFERS	1
+  struct file_dedupe_range_info {
+   __s64 fd; __u64 logical_offset; __u64 bytes_deduped; __s32 status; __u32 reserved;
+  };
+  struct file_dedupe_range {
+   __u64 logical_offset; __u64 length; __u16 dest_count; __u16 reserved1;
+   __u32 reserved2; struct file_dedupe_range_info info[0];
+  };
+  #define FIDEDUPERANGE _IOWR(BTRFS_IOCTL_MAGIC, 54, struct file_dedupe_range)
+  /* Static BTRFS header */
+ #else
+  #include <linux/fs.h>
+ #endif /* STATIC_BTRFS_H */
+#endif /* ENABLE_DEDUPE */
 
 #define JODY_HASH_WIDTH 32
 typedef uint32_t jodyhash_t;
@@ -321,7 +341,7 @@ static void nullptr(const char * restrict func)
 
 
 /* Jody Bruchon's fast hashing function
- * Copyright (C) 2014-2019 by Jody Bruchon <jody@jodybruchon.com>
+ * Copyright (C) 2014-2020 by Jody Bruchon <jody@jodybruchon.com>
  * Released under The MIT License
  */
 
@@ -619,18 +639,25 @@ static int getdirstats(const char * const restrict name,
  *  0 if all condition checks pass
  * -1 or 1 on compare result less/more
  * -2 on an absolute exclusion condition met
- *  2 on an absolute match condition met */
+ *  2 on an absolute match condition met
+ * -3 on exclusion due to isolation
+ * -4 on exlusion due to same filesystem
+ * -5 on exclusion due to permissions */
 static int check_conditions(const file_t * const restrict file1, const file_t * const restrict file2)
 {
   if (file1 == NULL || file2 == NULL || file1->d_name == NULL || file2->d_name == NULL) nullptr("check_conditions()");
 
+  /* Exclude files that are not the same size */
+  if (file1->size > file2->size) return -1;
+  if (file1->size < file2->size) return 1;
+
 #ifndef NO_USER_ORDER
   /* Exclude based on -I/--isolate */
-  if (ISFLAG(flags, F_ISOLATE) && (file1->user_order == file2->user_order)) return -1;
+  if (ISFLAG(flags, F_ISOLATE) && (file1->user_order == file2->user_order)) return -3;
 #endif /* NO_USER_ORDER */
 
   /* Exclude based on -1/--one-file-system */
-  if (ISFLAG(flags, F_ONEFS) && (file1->device != file2->device)) return -1;
+  if (ISFLAG(flags, F_ONEFS) && (file1->device != file2->device)) return -4;
 
    /* Exclude files by permissions if requested */
   if (ISFLAG(flags, F_PERMISSIONS) &&
@@ -640,7 +667,7 @@ static int check_conditions(const file_t * const restrict file1, const file_t *
           || file1->gid != file2->gid
 #endif
           )) {
-    return -1;
+    return -5;
   }
 
   /* Hard link and symlink + '-s' check */
@@ -651,10 +678,6 @@ static int check_conditions(const file_t * const restrict file1, const file_t *
   }
 #endif
 
-  /* Exclude files that are not the same size */
-  if (file1->size > file2->size) return -1;
-  if (file1->size < file2->size) return 1;
-
   /* Fall through: all checks passed */
   return 0;
 }
@@ -801,7 +824,7 @@ static unsigned int get_max_dupes(const file_t *files, unsigned int * const rest
 
 
 /* BTRFS deduplication of file blocks */
-#ifdef ENABLE_BTRFS
+#ifdef ENABLE_DEDUPE
 
 /* Message to append to BTRFS warnings based on write permissions */
 static const char *readonly_msg[] = {
@@ -812,8 +835,8 @@ static const char *readonly_msg[] = {
 static char *dedupeerrstr(int err) {
 
   tempname[sizeof(tempname)-1] = '\0';
-  if (err == BTRFS_SAME_DATA_DIFFERS) {
-    snprintf(tempname, sizeof(tempname), "BTRFS_SAME_DATA_DIFFERS (data modified in the meantime?)");
+  if (err == FILE_DEDUPE_RANGE_DIFFERS) {
+    snprintf(tempname, sizeof(tempname), "FILE_DEDUPE_RANGE_DIFFERS (data modified in the meantime?)");
     return tempname;
   } else if (err < 0) {
     return strerror(-err);
@@ -826,7 +849,7 @@ static char *dedupeerrstr(int err) {
 static void dedupefiles(file_t * restrict files)
 {
   struct utsname utsname;
-  struct btrfs_ioctl_same_args *same;
+  struct file_dedupe_range *same;
   char **dupe_filenames; /* maps to same->info indices */
 
   file_t *curfile;
@@ -854,8 +877,8 @@ static void dedupefiles(file_t * restrict files)
     fprintf(stderr, "Ask the program author to add this feature if you really need it. Exiting!\n");
     exit(EXIT_FAILURE);
   }
-  same = calloc(sizeof(struct btrfs_ioctl_same_args) +
-                sizeof(struct btrfs_ioctl_same_extent_info) * max_dupes, 1);
+  same = calloc(sizeof(struct file_dedupe_range) +
+                sizeof(struct file_dedupe_range_info) * max_dupes, 1);
   dupe_filenames = malloc(max_dupes * sizeof(char *));
   if (!same || !dupe_filenames) oom("dedupefiles() structures");
 
@@ -952,7 +975,7 @@ cleanup:
   free(dupe_filenames);
   return;
 }
-#endif /* ENABLE_BTRFS */
+#endif /* ENABLE_DEDUPE */
 
 
 /* Delete duplicate files automatically or interactively */
@@ -1749,6 +1772,7 @@ static void registerfile(filetree_t * restrict * const restrict nodeptr,
 static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict file)
 {
   int cmpresult = 0;
+  int cantmatch = 0;
   const jdupes_hash_t * restrict filehash;
 
   if (tree == NULL || file == NULL || tree->file == NULL || tree->file->d_name == NULL || file->d_name == NULL) nullptr("checkmatch()");
@@ -1767,6 +1791,12 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
   switch (cmpresult) {
     case 2: return &tree->file;  /* linked files + -H switch */
     case -2: return NULL;  /* linked files, no -H switch */
+    case -3:    /* user order */
+    case -4:    /* one filesystem */
+    case -5:    /* permissions */
+        cantmatch = 1;
+        cmpresult = 0;
+        break;
     default: break;
   }
 
@@ -1831,6 +1861,10 @@ static file_t **checkmatch(filetree_t * restrict tree, file_t * const restrict f
     }
   }
 
+  if( (cantmatch!=0) && (cmpresult==0) ) {
+    cmpresult = -1;
+  }
+
   if (cmpresult < 0) {
     if (tree->left != NULL) {
       return checkmatch(tree->left, file);
@@ -1993,8 +2027,8 @@ static inline void help_text(void)
   printf(" -0 --printnull   \toutput nulls instead of CR/LF (like 'find -print0')\n");
   printf(" -1 --one-file-system \tdo not match files on different filesystems/devices\n");
   printf(" -A --nohidden    \texclude hidden files from consideration\n");
-#ifdef ENABLE_BTRFS
-  printf(" -B --dedupe      \tsend matches to btrfs for block-level deduplication\n");
+#ifdef ENABLE_DEDUPE
+  printf(" -B --dedupe      \tsend matches to filesystem for block-level deduplication\n");
 #endif
   printf(" -d --delete      \tprompt user for files to preserve and delete all\n");
   printf("                  \tothers; important: under particular circumstances,\n");
@@ -2265,7 +2299,7 @@ int main(int argc, char **argv)
     case 'v':
     case 'V':
       printf("jdupes small stand-alone version (derived from v%s, %s)", VER, VERDATE);
-      printf("\nCopyright (C) 2015-2019 by Jody Bruchon <jody@jodybruchon.com>\n");
+      printf("\nCopyright (C) 2015-2020 by Jody Bruchon <jody@jodybruchon.com>\n");
       exit(EXIT_SUCCESS);
     case 'o':
       if (!strncasecmp("name", optarg, 5)) {
@@ -2278,7 +2312,7 @@ int main(int argc, char **argv)
       }
       break;
     case 'B':
-#ifdef ENABLE_BTRFS
+#ifdef ENABLE_DEDUPE
       SETFLAG(flags, F_DEDUPEFILES);
       /* btrfs will do the byte-for-byte check itself */
       SETFLAG(flags, F_QUICKCOMPARE);
@@ -2322,7 +2356,7 @@ int main(int argc, char **argv)
     exit(EXIT_FAILURE);
   }
 
-#ifdef ENABLE_BTRFS
+#ifdef ENABLE_DEDUPE
   if (ISFLAG(flags, F_CONSIDERHARDLINKS) && ISFLAG(flags, F_DEDUPEFILES))
     fprintf(stderr, "warning: option --dedupe overrides the behavior of --hardlinks\n");
 #endif
@@ -2466,9 +2500,9 @@ skip_file_scan:
 #ifndef NO_HARDLINKS
   if (ISFLAG(flags, F_HARDLINKFILES)) linkfiles(files, 1);
 #endif /* NO_HARDLINKS */
-#ifdef ENABLE_BTRFS
+#ifdef ENABLE_DEDUPE
   if (ISFLAG(flags, F_DEDUPEFILES)) dedupefiles(files);
-#endif /* ENABLE_BTRFS */
+#endif /* ENABLE_DEDUPE */
   if (ISFLAG(flags, F_PRINTMATCHES)) printmatches(files);
   if (ISFLAG(flags, F_PRINTJSON)) printjson(files, argc, argv);
   if (ISFLAG(flags, F_SUMMARIZEMATCHES)) {
