@@ -22,8 +22,6 @@ extern "C" {
  #include <windows.h>
  #include <io.h>
  #include "win_stat.h"
- #define S_ISREG WS_ISREG
- #define S_ISDIR WS_ISDIR
 #endif /* Win32 */
 
 #include <limits.h>
@@ -36,12 +34,6 @@ extern "C" {
 
 #include "xxhash.h"
 
-/* Optional btrfs support */
-#ifdef ENABLE_BTRFS
-#include <sys/ioctl.h>
-#include <linux/btrfs.h>
-#endif
-
 /* Set hash type (change this if swapping in a different hash function) */
  typedef XXH64_hash_t jdupes_hash_t;
 
@@ -73,7 +65,6 @@ extern "C" {
  #define WPATH_MAX 8192
  #define PATHBUF_SIZE WPATH_MAX
   typedef wchar_t wpath_t[WPATH_MAX];
-  extern wpath_t wname,wname2,wstr;
   extern int out_mode;
   extern int err_mode;
  #define M2W(a,b) MultiByteToWideChar(CP_UTF8, 0, a, -1, (LPWSTR)b, WPATH_MAX)
@@ -218,21 +209,23 @@ typedef struct _filetree {
 
 /* This gets used in many functions */
 #ifdef ON_WINDOWS
-extern struct winstat ws;
+ extern struct winstat s;
+ #define STAT win_stat
 #else
-extern struct stat s;
+ extern struct stat s;
+ #define STAT stat
 #endif
 
 
-/* -X exclusion parameter stack */
-struct exclude {
-  struct exclude *next;
+/* -X extended filter parameter stack */
+struct extfilter {
+  struct extfilter *next;
   unsigned int flags;
   int64_t size;
   char param[];
 };
 
-/* Exclude parameter flags */
+/* Extended filter parameter flags */
 #define X_DIR			0x00000001U
 #define X_SIZE_EQ		0x00000002U
 #define X_SIZE_GT		0x00000004U
@@ -249,13 +242,13 @@ struct exclude {
 #define XX_EXCL_DATA		0x0000000fU
 
 /* Exclude definition array */
-struct exclude_tags {
+struct extfilter_tags {
   const char * const tag;
   const uint32_t flags;
 };
 
-extern const struct exclude_tags exclude_tags[];
-extern struct exclude *exclude_head;
+extern const struct extfilter_tags extfilter_tags[];
+extern struct extfilter *extfilter_head;
 
 
 /* Suffix definitions (treat as case-insensitive) */
